/*======================================================================

Copyright (C) 1998 David Bateman -- dbateman@eng.uts.edu.au

This PCMCIA driver was written using dummy_cs as a starting point.
dummy_cs bares the following copyright
    
Copyright (C) 1998 David A. Hinds -- dhinds@hyper.stanford.edu

LICENSE
-------
This driver is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

It is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA

CHANGELOG
---------
Rev: 0.01 Date: 1998/02/20 19:54:23 (David Bateman)
  Started writing floppy driver based on skelton_cs (now dummy_cs)
Rev: 0.07 Date  1998/04/24 20:15:43 (David Bateman)
  Updated to more closely resemble dummy_cs in the pre 3.0.1 releases
  of linux PCMCIA package    
Rev: 0.08 Date  1998/05/14 02:11:16 (David Bateman)
  First pass at getting driver to work with PIO after receiving documentation
Rev: 0.30 Date  1998/05/17 13:55:34 (David Bateman)
  First working version. However there are problems with the DCL line flagging
  phantom disk changes causing some I/O errors. Workaround by forcing the
  FD_BROKEN_DCL option. Only a single FDC supported at the point, so if you
  plug in a FlashBuster with another FDC, you'll only see the FlashBuster. The
  floppy driver in the linux kernel MUST be built as a module!!
Rev: 0.40 Date 1998/05/25 01:12:48 (David Bateman)
  Japanese Mode 3 support. Included better workaround for the problem with
  the DCL. Inclusion of the PCMCIA floppy code in the kernel is now controlled
  with the CONFIG_BLK_DEV_FD_PCMCIA configuration option.
Rev 1.00 Date 1998/06/01 20:35:34 (David Bateman)
  Hopefully the final pre-release. If this release has no problems, will ask
  Y-E Data to relax NDA sufficiently to release the source code.
Rev 1.01 Date 1998/09/18 00:25:51 (David Bateman) (not released)
  Port the driver to the linux 2.1.122 kernel and update docs to reflect this
Rev 1.02 Date 1999/01/12 15:15:13 (David Bateman)
  Include better code to use a FlashBuster in the presence of another FDC. Add
  a troubleshooting section to the README.
======================================================================*/

#include <pcmcia/config.h>
#include <pcmcia/k_compat.h>

#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/ptrace.h>
#include <linux/malloc.h>
#include <linux/string.h>
#include <linux/timer.h>
#ifndef CONFIG_BLK_DEV_FD_PCMCIA
#define CONFIG_BLK_DEV_FD_PCMCIA
#endif
#include <linux/fdreg.h>

#include <linux/ioport.h>
#include <linux/major.h>
#include <asm/io.h>
#include <asm/system.h>

#include <pcmcia/version.h>
#include <pcmcia/cs_types.h>
#include <pcmcia/cs.h>
#include <pcmcia/cistpl.h>
#include <pcmcia/ds.h>
#include <pcmcia/cisreg.h>

#ifdef PCMCIA_DEBUG
static int pc_debug = PCMCIA_DEBUG;
MODULE_PARM(pc_debug, "i");
#define DEBUG(n, args...) if (pc_debug>(n)) printk(KERN_DEBUG args);
static char *version =
"floppy_cs.c 1.02 Date 1999/01/12 15:15:13 (David Bateman)";
#else
#define DEBUG(n, args...)
#endif

/*====================================================================*/

/* Parameters that can be set with 'insmod' */

/* Bit map of interrupts to choose from */
static u_long irq_mask = 0xdeb8;
static int irq_list[16] = { -1 };
static int mode3 = 0;
static int iobase = 0;

MODULE_PARM(irq_mask, "i");
MODULE_PARM(irq_list, "1-4i");
MODULE_PARM(mode3, "i");
MODULE_PARM(iobase, "i");

/*====================================================================*/

typedef struct floppy_info_t {
    dev_node_t	node;
    unsigned int flags;
} floppy_info_t;


static void floppy_config(dev_link_t *link);
static void floppy_release(u_long arg);
static int floppy_event(event_t event, int priority,
			event_callback_args_t *args);

static dev_info_t dev_info = "floppy_cs";

static dev_link_t *floppy_attach(void);
static void floppy_detach(dev_link_t *);

static dev_link_t *dev_list = NULL;

/*====================================================================*/

static void cs_error(client_handle_t handle, int func, int ret)
{
    error_info_t err = { func, ret };
    CardServices(ReportError, handle, &err);
}

/*======================================================================

    floppy_attach() creates an "instance" of the driver, allocating
    local data structures for one device.  The device is registered
    with Card Services.

======================================================================*/

static dev_link_t *floppy_attach(void)
{
    client_reg_t client_reg;
    dev_link_t *link;
    int i, ret;
    
    DEBUG(0, "floppy_attach()\n");

    /* Create new floppy device */
    link = kmalloc(sizeof(struct dev_link_t), GFP_KERNEL);
    memset(link, 0, sizeof(struct dev_link_t));
    link->release.function = &floppy_release;
    link->release.data = (u_long)link;

    link->io.NumPorts1 = 8;
    link->io.Attributes1 = IO_DATA_PATH_WIDTH_8;
    link->io.NumPorts2 = 0;
    link->io.IOAddrLines = 3;

    link->irq.Attributes = IRQ_TYPE_EXCLUSIVE;
    link->irq.IRQInfo1 = IRQ_INFO2_VALID|IRQ_LEVEL_ID;

    if (irq_list[0] == -1)
	link->irq.IRQInfo2 = irq_mask;
    else {
	link->irq.IRQInfo2 = 0;
	for (i = 0; i < 16; i++)
	    link->irq.IRQInfo2 |= 1 << irq_list[i];
    }
    
    link->conf.Attributes = CONF_ENABLE_IRQ;

    link->conf.Vcc = 50;
    link->conf.IntType = INT_MEMORY_AND_IO;
    link->conf.ConfigIndex = 1;
    link->conf.Present = PRESENT_OPTION;
    link->priv = kmalloc(sizeof(struct floppy_info_t), GFP_KERNEL);
    memset(link->priv, 0, sizeof(struct floppy_info_t));
    
    /* Register with Card Services */
    link->next = dev_list;
    dev_list = link;
    client_reg.dev_info = &dev_info;
    client_reg.Attributes = INFO_IO_CLIENT | INFO_CARD_SHARE;
    client_reg.EventMask =
	CS_EVENT_CARD_INSERTION | CS_EVENT_CARD_REMOVAL |
	CS_EVENT_RESET_PHYSICAL | CS_EVENT_CARD_RESET |
	CS_EVENT_PM_SUSPEND | CS_EVENT_PM_RESUME;
    client_reg.event_handler = &floppy_event;
    client_reg.Version = 0x0210;
    client_reg.event_callback_args.client_data = link;
    ret = CardServices(RegisterClient, &link->handle, &client_reg);
    if (ret != CS_SUCCESS) {
	cs_error(link->handle, RegisterClient, ret);
	floppy_detach(link);
	return NULL;
    }
    
    return link;
} /* floppy_attach */

/*======================================================================

    This deletes a driver "instance".  The device is de-registered
    with Card Services.  If it has been released, all local data
    structures are freed.  Otherwise, the structures will be freed
    when the device is released.

======================================================================*/

static void floppy_detach(dev_link_t *link)
{
    dev_link_t **linkp;

    DEBUG(0, "floppy_detach(0x%p)\n", link);
    
    /* Locate device structure */
    for (linkp = &dev_list; *linkp; linkp = &(*linkp)->next)
	if (*linkp == link) break;
    if (*linkp == NULL)
	return;

    /*
       If the device is currently configured and active, we won't
       actually delete it yet.  Instead, it is marked so that when
       the release() function is called, that will trigger a proper
       detach().
    */
    if (link->state & DEV_CONFIG) {
	DEBUG(0, "floppy_cs: detach postponed, '%s' still locked\n",
	      link->dev->dev_name);
	link->state |= DEV_STALE_LINK;
	return;
    }

    /* Break the link with Card Services */
    if (link->handle)
	CardServices(DeregisterClient, link->handle);
    
    /* Unlink device structure, free pieces */
    *linkp = link->next;
    if (link->priv) {
	kfree_s(link->priv, sizeof(floppy_info_t));
    }
    kfree_s(link, sizeof(struct dev_link_t));
    
} /* floppy_detach */

/*======================================================================

    floppy_config() is scheduled to run after a CARD_INSERTION event
    is received, to configure the PCMCIA socket, and to make the
    floppy device available to the system.

======================================================================*/

#define CS_CHECK(fn, args...) \
while ((last_ret=CardServices(last_fn=(fn),args))!=0) goto cs_failed

#define CFG_CHECK(fn, args...) \
if (CardServices(fn, args) != 0) goto next_entry

void floppy_config(dev_link_t *link)
{
    client_handle_t handle;
    tuple_t tuple;
    cisparse_t parse;
    floppy_info_t *dev;
    int i, last_fn, last_ret;
    u_char buf[256];

    handle = link->handle;
    dev = link->priv;

    DEBUG(0, "floppy_config(0x%p)\n", link);

    if (mode3) dev->flags |= PCMCIA_MODE3;

    tuple.DesiredTuple = CISTPL_CONFIG;
    tuple.Attributes = 0;
    tuple.TupleData = buf;
    tuple.TupleDataMax = 255;
    tuple.TupleOffset = 0;
    CS_CHECK(GetFirstTuple, handle, &tuple);
    CS_CHECK(GetTupleData, handle, &tuple);
    CS_CHECK(ParseTuple, handle, &tuple, &parse);
    link->conf.ConfigBase = parse.config.base;
    link->conf.Present = parse.config.rmask[0];

    /* Configure card */
    link->state |= DEV_CONFIG;

    tuple.DesiredTuple = CISTPL_CFTABLE_ENTRY;
    CS_CHECK(GetFirstTuple, handle, &tuple);
    while (1) {
	cistpl_cftable_entry_t dflt = { 0 };
	cistpl_cftable_entry_t *cfg = &(parse.cftable_entry);
        CS_CHECK(GetTupleData, handle, &tuple);
	CS_CHECK(ParseTuple, handle, &tuple, &parse);

	if (cfg->index == 0) goto next_entry;
	link->conf.ConfigIndex = cfg->index;
	
	/* Use power settings for Vcc and Vpp if present */
	/*  Note that the CIS values need to be rescaled */
	if (cfg->vcc.present & (1<<CISTPL_POWER_VNOM))
	    link->conf.Vcc = cfg->vcc.param[CISTPL_POWER_VNOM]/10000;
	else if (dflt.vcc.present & (1<<CISTPL_POWER_VNOM))
	    link->conf.Vcc = dflt.vcc.param[CISTPL_POWER_VNOM]/10000;
	    
	if (cfg->vpp1.present & (1<<CISTPL_POWER_VNOM))
	    link->conf.Vpp1 = link->conf.Vpp2 =
		cfg->vpp1.param[CISTPL_POWER_VNOM]/10000;
	else if (dflt.vpp1.present & (1<<CISTPL_POWER_VNOM))
	    link->conf.Vpp1 = link->conf.Vpp2 =
		dflt.vpp1.param[CISTPL_POWER_VNOM]/10000;
	
	/* Do we need to allocate an interrupt? */
	if (cfg->irq.IRQInfo1 || dflt.irq.IRQInfo1)
	    link->conf.Attributes |= CONF_ENABLE_IRQ;
	
	if (cfg->io.nwin > 0) {
	    if (iobase) {
	        if (cfg->io.win[0].base == iobase) {
		    link->io.BasePort1 = cfg->io.win[0].base;
		    CFG_CHECK(RequestIO, link->handle, &link->io);

		    /* We're done, bug out */
		    break;
		} 
	    } else if (cfg->io.win[0].base > 0x2ff) {
	        /* Sanity checks. Avoid address below 0x300 due to possible */
	        /* conflict with APM and hard drive controllers             */
	        link->io.BasePort1 = cfg->io.win[0].base;
		CFG_CHECK(RequestIO, link->handle, &link->io);

		/* We're done, bug out */
		break;
	    }
	}
	
    next_entry:
	if (cfg->flags & CISTPL_CFTABLE_DEFAULT) dflt = *cfg;
	CS_CHECK(GetNextTuple, handle, &tuple);
    }
	
    /*
      Now allocate an interrupt line.  Note that this does not
      actually assign a handler to the interrupt.
      */
    if (link->conf.Attributes & CONF_ENABLE_IRQ)
	CS_CHECK(RequestIRQ, link->handle, &link->irq);
	
    /*
      This actually configures the PCMCIA socket -- setting up
      the I/O windows and the interrupt mapping.
      */
    CS_CHECK(RequestConfiguration, link->handle, &link->conf);

    release_region(link->io.BasePort1, link->io.NumPorts1);
    /* Now initialize the floppy driver */
    i = floppy_register(link->io.BasePort1, link->irq.AssignedIRQ,
			&(dev->flags));
    if (i != 0) {
      printk(KERN_INFO "floppy_cs: Register PCMCIA floppy IRQ %d, Base 0x%X, returned error %d\n", link->irq.AssignedIRQ, link->io.BasePort1, i);
      goto failed;
    }

    /* One day perhaps this drive might support coexisting with another FDC */
    switch (dev->flags & PCMCIA_FDC1) {
	case 1:
	  sprintf(dev->node.dev_name, "fd4");
	  break;
	default:
	case 0:
	  sprintf(dev->node.dev_name, "fd0");
	  break;
    }
    dev->node.major = FLOPPY_MAJOR;
    dev->node.minor = (dev->flags & PCMCIA_FDC1) * 128;
    link->dev = &dev->node;
    
    link->state &= ~DEV_CONFIG_PENDING;

    DEBUG(0, "floppy device loaded IRQ %d, Base 0x%X\n",
	  (unsigned int)link->irq.AssignedIRQ,
	  (unsigned int)link->io.BasePort1);
    return;

cs_failed:
    cs_error(link->handle, last_fn, last_ret);
failed:
    floppy_release((u_long)link);

} /* floppy_config */

/*======================================================================

    After a card is removed, floppy_release() will unregister the
    device, and release the PCMCIA configuration.  If the device is
    still open, this will be postponed until it is closed.
    
======================================================================*/

void floppy_release(u_long arg)
{
    dev_link_t *link = (dev_link_t *)arg;

    DEBUG(0, "floppy_release(0x%p)\n", link);

    /*
       If the device is currently in use, we won't release until it
       is actually closed.
    */
    if (link->open) {
        DEBUG(0, "floppy_cs: release postponed, '%s' still open\n",
	    link->dev->dev_name);
	link->state |= DEV_STALE_CONFIG;
	return;
    }

    /* Unlink the device chain */
    link->dev = NULL;

    /* Clean up the floppy device driver */
    floppy_unregister(link->io.BasePort1);

    /* Don't bother checking to see if these succeed or not */
    if (link->win)
	CardServices(ReleaseWindow, link->win);
    CardServices(ReleaseConfiguration, link->handle);
    if (link->io.NumPorts1)
	CardServices(ReleaseIO, link->handle, &link->io); 
    if (link->irq.AssignedIRQ)
	CardServices(ReleaseIRQ, link->handle, &link->irq);
    link->state &= ~DEV_CONFIG;
    
    if (link->state & DEV_STALE_LINK)
	floppy_detach(link);
    
} /* floppy_release */

/*======================================================================

    The card status event handler.  Mostly, this schedules other
    stuff to run after an event is received.  A CARD_REMOVAL event
    also sets some flags to discourage the floppy drivers from
    talking to the ports.
    
======================================================================*/

int floppy_event(event_t event, int priority,
		event_callback_args_t *args)
{
    dev_link_t *link = args->client_data;
    floppy_info_t *dev;

    
    DEBUG(1, "floppy_event()\n");
    
    dev = link->priv;

    switch (event) {
    case CS_EVENT_CARD_REMOVAL:
	link->state &= ~DEV_PRESENT;
	/* Turn off mode 3 to avoid interrupt from FDC */
	if (dev->flags && PCMCIA_MODE3)
	    fdc_set_mode3(((dev->flags & PCMCIA_FDC1) ? 1 : 0), 0);
	if (link->state & DEV_CONFIG) {
	    link->release.expires = RUN_AT(HZ/20);
	    link->state |= DEV_RELEASE_PENDING;
	    add_timer(&link->release);
	}
	break;
    case CS_EVENT_CARD_INSERTION:
	link->state |= DEV_PRESENT | DEV_CONFIG_PENDING;
	floppy_config(link);
	break;
    case CS_EVENT_PM_SUSPEND:
	link->state |= DEV_SUSPEND;
	/* Fall through... */
    case CS_EVENT_RESET_PHYSICAL:
	/* Turn off mode 3 to avoid interrupt from FDC */
	if (dev->flags && PCMCIA_MODE3)
	    fdc_set_mode3(((dev->flags & PCMCIA_FDC1) ? 1 : 0), 0);
	if (link->state & DEV_CONFIG)
	    CardServices(ReleaseConfiguration, link->handle);
	break;
    case CS_EVENT_PM_RESUME:
	link->state &= ~DEV_SUSPEND;
	/* Fall through... */
    case CS_EVENT_CARD_RESET:
	if (link->state & DEV_CONFIG)
	    CardServices(RequestConfiguration, link->handle, &link->conf);
	/* Should turn mode 3 on here, but causes interrupt. Do in fdc_dtr */
	break;
    }
    return 0;
} /* floppy_event */

/*====================================================================*/

int init_module(void)
{
    servinfo_t serv;
    DEBUG(0, "%s\n", version);
    CardServices(GetCardServicesInfo, &serv);
    if (serv.Revision != CS_RELEASE_CODE) {
	printk(KERN_NOTICE "floppy_cs: Card Services release "
	       "does not match!\n");
	return -1;
    }
    register_pcmcia_driver(&dev_info, &floppy_attach, &floppy_detach);
    return 0;
}

void cleanup_module(void)
{
    DEBUG(0, "floppy_cs: unloading\n");
    unregister_pcmcia_driver(&dev_info);
    while (dev_list != NULL) {
	if (dev_list->state & DEV_CONFIG)
	    floppy_release((u_long)dev_list);
	floppy_detach(dev_list);
    }
}
